/*
 *  Copyright (C) 2000  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */

#include <ctype.h>
#include <string.h>
#include "digicons.h"
#include "digitype.h"
#include "output.h"
#include "genlib.h"
#ifdef _LINUX_
#include "linux.h"
#endif

/*
 * Compare call with rule-call.
 *
 * The rule-call can contain special characters:
 *
 * '@' match any letter
 * '!' match any digit
 * '?' match any character
 * '*' match 0 or more characters, any character
 * 
 */
short match_call(const char *call, const char *rule_call)
{
    const char *p, *q;

    p = call;
    q = rule_call;

    while(*q != '\0')
    {
        if(*q == '*')
        {
            /* advance to next character */
            q++;

            /* the following constuct is to match rule-calls of either */
            /* "*" or "*-<SSID>", this enables to-call selection on    */
            /* SSID to enable digipeating on SSID destination.         */

            /* now proceed until the character '*q' matches with */
            /* '*p' again                                        */
            while(*p != *q)
            {
                if(*p == '\0')
                {
                    /* *p finished while *q is not, no match */
                    return 0;
                }
                p++;
            }

            /* *p and *q are in sync now */
            if(*q == '\0')
            {
                /* the end is reached on both strings, match is completed */
                return 1;
            }
        }
        else
        {
            if(*p == '\0')
            {
                /* p is at the end, but q is not */
                return 0;
            }

            switch(*q) {
            case '?':
                    /* okay, whatever it is... */
                    break;
            case '@':
                    /* when not alphanumeric, quit */
                    if(!isalpha(*q))
                        return 0;
                    break;
            case '!':
                    /* when not a digit, quit */
                    if(!isdigit(*q))
                        return 0;
                    break;
            default:
                    if(*p != *q)
                    {
                        /* not the same, quit */
                        return 0;
                    }
                    break;
            }
            p++;
            q++;
        }
    }

    /* check if p is also finished */
    if(*p == '\0')
        return 1;

    return 0;
}

/*
 * Check if ports parameter contains correct characters
 *
 * Returns: 0 if not okay
 *          1 if okay for portnumbers and "all"
 *
 * Note: ports string will be converted to lowercase
 */
short check_ports(char *ports, const char *purpose, const char *place,
                    short line, porttype_t porttype)
{
    strlwr(ports);
    if(strlen(ports) > OUTPORTS_LENGTH)
    {
        say("%s (%s) to long on %s: at line %d\r\n",
                purpose, ports, place, line);
        return 0;
    }

    if(strcmp(ports, "all") == 0)
    {
        /* "all" is accepted */
        return 1;
    }

    if((porttype == INCL_ALLBUT) && (strcmp(ports, "allbut") == 0))
    {
        /* "allbut" is accepted */
        return 1;
    }

    if(strspn(ports, "0123456789,") == strlen(ports))
    {
        /* no illegal characters, accepted */
        return 1;
    }

    say("%s (%s) has illegal format on %s: at line %d\r\n",
            purpose, ports, place, line);
    return 0;
}

/*
 * Check if the input string is a valid call
 *
 * Returns: 0 if not okay
 *          1 if okay
 */
short is_call(const char *c)
{
    short n;

    if(*c == '\0')
    {
        /* Missing call */
        return 0;
    }

    /* call-size counter to 0 */
    n = 0;

    do
    {
        /* count call size */
        n++;

        if(n == 7)
        {
            /* Illegal call length */
            return 0;
        }

        /*
         * Before the dash can be either a digit
         * or an uppercase alpha character
         */
        if(!isdigit(*c) && !isupper(*c))
        {
            /* Illegal character in call */
            return 0;
        }

        /* advance to next character */
        c++;

        /* if end of string, checking completed successfully */
        if(*c == '\0')
        {
            /* okay, call without SSID */
            return 1;
        }
    } while(*c != '-');

    /* if we end up here then a '-' was found */

    /* advance to first character of SSID */
    c++;

    if(*c == '\0')
    {
        /* Missing SSID */
        return 0;
    }

    if(*c == '1')
    {
        /* can be a 2 digit SSID */

        /* next character can be '\0' */
        c++;

        if(*c == '\0')
        {
            /* SSID is '1', okay */
            return 1;
        }

        /* remaining character shall be 0..5 */
        if((*c < '0') || (*c > '5'))
        {
            /* Impossible digit value */
            return 0;
        }
    }
    else
    {
        /* single digit SSID */
        if(!isdigit(*c))
        {
            /* Illegal character in SSID */
            return 0;
        }
    }

    /* next character shall be '\0' */
    c++;

    if(*c != '\0')
    {
        /* characters following the SSID */
        return 0;
    }
    else
    {
        /* checking completed successfully */
        return 1;
    }
}

/*
 * Determine the number of hops from the originator
 *
 * Returns: 0 if station has been heard directly
 *          n the number of hops to get here
 */
short number_of_hops(uidata_t *uidata_p)
{
    short hops;
    short i;

    hops = 0;
   
    if (uidata_p->distance != LOCAL)
    {
        for (i = 0; i < uidata_p->ndigi; i++)
        {
            if ((uidata_p->digi_flags[i] & H_FLAG) != 0)
            {
                hops++;
            }
        }
    }
    return hops;
}
