/*
 *  Copyright (C) 2000  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/stat.h>
#ifdef _LINUX_
#include "linux.h"
#endif /* _LINUX_ */
#include "version.h"
#include "digicons.h"
#include "digitype.h"
#include "genlib.h"
#include "mac_if.h"
#include "read_ini.h"
#include "message.h"
#include "keep.h"
#include "call.h"
#include "timer.h"
#include "output.h"
#include "mheard.h"
#include "send.h"

/*
 * Linked lists containing the 'message_path' information per port
 */
static message_path_t *port_message_path_p[MAX_PORTS] = { 
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };

static char        message_id[MAX_NR_OF_IDS];
static message_t  *msg_list_p;
static time_t      boot_time;
static exit_info_t exit_info;
static query_t    *digi_query_cur_p;
static query_t    *digi_query = NULL;
static struct stat digi_query_stat;
static short       keyboard_command;

/*
 * Add message_path command from the .ini file to the internal administration.
 *
 * message_path: 1,2,3,4,5,6,7,8 WIDE,TRACE6-6
 *              ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'message_path' token is already handled. The tokenizer is now at the
 * point in the string indicated by '^' above (the ':' was changed to
 * '\0' by 'strtok').
 */
void add_message_path(short ini_ports)
{
    message_path_t message_path;
    message_path_t *message_path_p;
    char   *tmp;
    char   *path;
    char    output[OUTPORTS_LENGTH];
    short   port;

    tmp = strtok(NULL, " \t\r\n");
    if(tmp == NULL)
    {
        say("Missing to-port on message_path: at line %d\r\n", linecount);
        return;
    }

    if(check_ports(tmp, "To-port", "message_path:", 
                   linecount, EXCL_ALLBUT) == 0)
    {
        return;
    }

    strcpy(output, tmp);

    message_path.path = NULL;

    path = strtok(NULL, "\r\n");

    if(path != NULL)
    {
        strupr(path);

        message_path.path = strdup(path);

        if(message_path.path == NULL)
        {
            say("Out of memory on message_path: at line %d\r\n", linecount);
            return;
        }

        /*
         * Check for illegal digipeaters (too long)
         */
        tmp = strtok(path, " ,\t\r\n");
        while(tmp != NULL)
        {
            if(strlen(tmp) > 9)
            {
                say("Argument (%s) to long on message_path: at line %d\r\n",
                        tmp, linecount);

                free(message_path.path);
                return;
            }
            tmp = strtok(NULL, " ,\t\r\n");
        }
    }

    /*
     * Store the message_path in the port_message_path list for each port
     */
    if(strcmp(output,"all") == 0)
    {
        for(port = 0; port < ini_ports; port++)
        {
            /*
             * Allocate dynamic storage for this entry
             */
            message_path_p = (message_path_t*) malloc(sizeof(message_path_t));

            /*
             * if out of memory then ignore
             */
            if(message_path_p == NULL)
            {
                say("Out of memory while handling message_path: "
                        "line at line %d\r\n", linecount);
                return;
            }

            /*
             * Duplicate data
             */
            memcpy(message_path_p, &message_path, sizeof(message_path_t));

            /*
             * Add to the message_path list for this port
             */
            message_path_p->next = port_message_path_p[port];
            port_message_path_p[port] = message_path_p;
        }
    }
    else
    {
        /* handle each port */
        tmp = strtok(output, ",");
        while(tmp != NULL)
        {
            port = atoi(tmp);
            if(port < 1 || port > ini_ports)
            {
                say("To-port out of range (1-%d) on message_path: at "
                        "line %d\r\n", ini_ports, linecount);
                return;
            }
            port--; /* make base 0 numbers */

            /*
             * Allocate dynamic storage for this entry
             */
            message_path_p = (message_path_t*) malloc(sizeof(message_path_t));

            /*
             * if out of memory then ignore
             */
            if(message_path_p == NULL)
            {
                say("Out of memory while handling message_path: "
                        "line at line %d\r\n", linecount);
                return;
            }
            
            /*
             * Duplicate data
             */
            memcpy(message_path_p, &message_path, sizeof(message_path_t));

            /*
             * Add to the message_path list for this port
             */
            message_path_p->next = port_message_path_p[port];
            port_message_path_p[port] = message_path_p;

            tmp = strtok(NULL, ",");
        }
    }
}

/*
 * Read in the message file
 */
static void read_queries(void)
{
    FILE      *fd;
    struct tm *t;
    char       query[MAX_MESSAGE_LENGTH + 1];
    char       answer[MAX_DATA_LENGTH + 1];
    char       buf[MAX_DATA_LENGTH + 1];
    char       tmp[40];
    char      *p;
    char      *q;
    short      ports;
    query_t   *query_p;
    query_t   *query_next_p;

    digi_query = NULL;

    if(stat(digi_message_file, &digi_query_stat) != 0)
    {
        say("Can't find message query/response file \"%s\"",
                digi_message_file);
        return;
    }

    fd = fopen(digi_message_file, "r");

    if(fd == NULL)
    {
        say("Can't read message query/response file \"%s\"",
                digi_message_file);
        return;
    }

    while(fgets(buf, MAX_DATA_LENGTH, fd) != NULL)
    {
        /* truncate at maximum allowable length */
        buf[MAX_MESSAGE_LENGTH] = '\0';

        /* truncate at the first CR or LF character, if any */
        p = strchr(buf, '\r');
        if(p != NULL) *p = '\0';
        p = strchr(buf, '\n');
        if(p != NULL) *p = '\0';

        /* ignore comment-lines */
        if(buf[0] ==  '#')
            continue;

        /* ignore empty-lines */
        if(strlen(buf) == 0)
            continue;

        if(buf[0] ==  '?')
        {
            strcpy(query, buf);
            strlwr(query);
            answer[0] = '\0';
        }
        else
        {
            p = buf;

            if(*p == '\\')
                p++;

            q = answer;

            while(*p != '\0')
            {
                if(*p == '%')
                {
                    p++;
                    switch(*p)
                    {
                    case 'd':
                        strcat(answer, digi_digi_call);
                        p++;
                        break;
                    case 'o':
                        strcat(answer, digi_owner());
                        p++;
                        break;
                    case 'v':
                        sprintf(tmp, "%s: %s - %s", VERSION,__DATE__,__TIME__);
                        strcat(answer, tmp);
                        p++;
                        break;
                    case 'b':
                        tzset();
                        t = localtime(&boot_time);
                        strftime(tmp, 39, "%b %d %Y - %H:%M:%S", t);
                        strcat(answer, tmp);
                        p++;
                        break;
                    case 'p':
                        ports = mac_ports();
                        sprintf(tmp, "%d", ports);
                        strcat(answer, tmp);
                        p++;
                        break;
                    default:
                        *q++ = '%';
                        *q = '\0';
                        break;
                    }
                    q = &(answer[strlen(answer)]);
                }
                else
                {
                    *q++ = *p++;
                    *q = '\0';
                }
                if(strlen(answer) >= MAX_MESSAGE_LENGTH)
                    break;
            }

            answer[MAX_MESSAGE_LENGTH] = '\0';

            query_p = malloc(sizeof(query_t));
            if(query_p != NULL)
            {
                strcpy(query_p->query, query);
                strcpy(query_p->answer, answer);
                
                query_p->next = digi_query;
                digi_query = query_p;
            }
            else
            {
                say("Out of memory while reading message query/answers\n");
            }
        }
    }
    
    fclose(fd);

    /* Queries are now stored in reverse order, swap the order */
    query_p = digi_query;
    digi_query = NULL;
    while(query_p != NULL)
    {
        query_next_p = query_p->next;
        
        query_p->next = digi_query;
        digi_query = query_p;

        query_p = query_next_p;
    }

    return;
}

/*
 * Erase the queries
 */
static void erase_queries(void)
{
    query_t  *query_p;
    query_t  *query_next_p;

    /* set query_p to the start of the list */
    query_p = digi_query;

    /* This will erase the queries */
    digi_query = NULL;

    /* Free all allocated memory */
    while(query_p != NULL)
    {
        query_next_p = query_p->next;
        
        free(query_p);

        query_p = query_next_p;
    }

    return;
}

/*
 * Check if the queries are changed and need to be re-read
 */
static void check_queries(void)
{
    struct stat s;

    if(stat(digi_message_file, &s) != 0)
    {
        /* Can't find, can't update, go with what we have */
        return;
    }

    if(digi_query_stat.st_mtime == s.st_mtime)
    {
        /* same modification time, not changed, go with what we have */
        return;
    }

    /* File changed */

    /* Remove the old message query/responses */
    erase_queries();

    /* Read the new message query/responses from the digi_message_file */
    read_queries();

    return;
}

/*
 * Transmit message
 */
static void transmit_message(message_t *message_p)
{
    uidata_t        uidata;
    message_path_t *message_path_p;
    short           ndigi;
    char           *tmp;
    char           *path;
    char            output[MAX_ITOA_LENGTH];

    /* prepare transmission */
    uidata.port = message_p->port;

    /* transmit remotely */
    uidata.distance = REMOTE;

    /* primitive UI with no poll bit */
    uidata.primitive = 0x03;

    /* PID = 0xF0, normal AX25*/
    uidata.pid = 0xF0;

    /* fill in originator */
    strcpy(uidata.originator, digi_digi_call);
    uidata.orig_flags = (char) (RR_FLAG | C_FLAG);

    /* fill in destination */
    strcpy(uidata.destination, digi_digi_dest);
    uidata.dest_flags = RR_FLAG;

    message_path_p = port_message_path_p[uidata.port];

    /* handle all the paths for this port */
    while(message_path_p != NULL)
    {
        ndigi=0;
        if(message_path_p->path != NULL)
        {
            /* make copy of digipeater path for strtok */
            path = strdup(message_path_p->path);
            if(path == NULL)
            {
                /* no memory for transmission */
                return;
            }

            /* add digipeater path */
            tmp = strtok(path, " ,\t\r\n");
            while(tmp != NULL)
            {
                if(ndigi < MAX_DIGI)
                {
                    strcpy(uidata.digipeater[ndigi], tmp);
                    uidata.digi_flags[ndigi] = RR_FLAG;
                    ndigi++;
                }
                tmp = strtok(NULL, " ,\t\r\n");
            }

            /* don't need "path" anymore, free memory */
            free(path);
        }
        uidata.ndigi = ndigi;

        /* a shutdown message should not have any paths */
        if(exit_info.state == CLOSING_DOWN)
        {
            if(message_p->id == exit_info.id)
            {
                /* 
                 * this is the shutdown message, should have no digis
                 * eliminate the digis by setting the count to 0
                 */
                uidata.ndigi = 0;
            }
        }

        /* set the data to transmit */
        strcpy(uidata.data, message_p->message);
        uidata.size = strlen(uidata.data);

        /* avoid digipeating my own data */
        keep_old_data(&uidata);

        /* convert port to string for data dump */
        itoa(message_p->port, output, 10);

        vsay("Message transmission:\r\n");

        /* show what we will transmit */
        dump_uidata_to(output, &uidata);

        /* transmit */
        put_uidata(&uidata);

        /* next path */
        message_path_p = message_path_p->next;
    }
}

/*
 * Retrieve message ID, as close following the last returned number
 *
 * returns: 0-99 message id to use with the next message (allocated)
 *          -1 if there are no free message id's anymore (all waiting for ack)
 */
static short get_message_id(void)
{
    static short last = -1;
    short        i;

    if(last == -1)
    {
        /* first time to retrieve an ID, start somewhere between 0 and 99 */
        last = (short) (((unsigned short) time(NULL)) % MAX_NR_OF_IDS);
    }

    /* try to find an unallocated message id */
    for(i = 0; i < MAX_NR_OF_IDS; i++)
    {
        last = (last + 1) % MAX_NR_OF_IDS;
        if(message_id[last] == 0)
        {
            /* allocate */
            message_id[last] = 1;
            return last;
        }
    }
    return -1;
}

/*
 * Clear a message from the message-queue
 */
static void remove_message(short id)
{
    message_t *message_p;

    message_p = msg_list_p;
    while(message_p != NULL)
    {
        if(message_p->id == id)
        {
            /* mark for destruction */
            message_p->count = 0;

            /* expire the timer */
            clear_timer(&(message_p->timer));

            return;
        }
        message_p = message_p->next;
    }
    return;
}

/*
 * send an ack on a received message
 */
static void send_ack(char *id, uidata_t *uidata_p)
{
    message_t *message_p;

    /* create a new message */
    message_p = malloc(sizeof(message_t));

    if(message_p == NULL)
    {
        /* no memory, more luck next time! */
        return;
    }

    /* make message header */
    sprintf(message_p->message, ":         :ack%s", id);

    /* patch in destination addres */
    strncpy(&(message_p->message[1]), uidata_p->originator,
                                strlen(uidata_p->originator));

    /* send back wher the message came from */
    message_p->port = uidata_p->port;

    /* no message id */
    message_p->id = -1;

    /* interval of 4 seconds to start with */
    message_p->interval = 4;

    /* send the ack only 1 time for each reception */
    message_p->count = 1;

    /* start interval-time seconds after reception */
    start_timer(&(message_p->timer), message_p->interval * TM_SEC);

    /* add to msg_list_p, new message in front */
    message_p->next = msg_list_p;
    msg_list_p = message_p;
}

/*
 * send a message to originator indicated in the uidata_p frame
 *
 * transmission is not done here, but the message is added to a
 * list of messages which are scheduled for transmission (until acked or
 * the count value reaches 0)
 */
static void send_message(ack_t need_ack, short number, char *message,
                                                    uidata_t *uidata_p)
{
    message_t *message_p;
    char       output[MAX_ITOA_LENGTH];

    if(keyboard_command == 1)
    {
        say("Reply %d: %s\r\n", number, message);
        return;
    }

    /* create a new message */
    message_p = malloc(sizeof(message_t));

    if(message_p == NULL)
    {
        /* no memory, more luck next time! */
        return;
    }

    /* make message header */
    strcpy(message_p->message, ":         :");

    /* patch in destination addres */
    strncpy(&(message_p->message[1]), uidata_p->originator,
                                strlen(uidata_p->originator));
    
    /* 
     * Copy message in. Maximum length of a message is = MAX_MESSAGE_LENGTH
     * bytes.
     */
    strncpy(&(message_p->message[MSG_PREFIX_LENGTH]), 
            message, MAX_MESSAGE_LENGTH);

    /* make shure there is a null terminator at the end */
    message_p->message[MSG_PREFIX_LENGTH + MAX_MESSAGE_LENGTH] = '\0';

    /* send back where the message came from */
    message_p->port = uidata_p->port;

    if(need_ack == NOACK)
    {
        /* no message id */
        message_p->id = -1;
    }
    else /* ACK or SHUTDOWN */
    {
        /* get message id */
        message_p->id = get_message_id();

        if((need_ack == SHUTDOWN) && (message_p->id != -1))
        {
            /* cleanup old shutdown message if present */
            if(exit_info.state == CLOSING_DOWN)
            {
                /*
                 * We were already closing down, remove old pending
                 * shutdown message
                 */
                remove_message(exit_info.id);
            }

            /* remember the new id */
            exit_info.id = message_p->id;

            /* mark that we are closing down now */
            exit_info.state = CLOSING_DOWN;
        }
    }

    /* add indicator to message when id is used */
    if(message_p->id != -1)
    {
        /* add indicator to the message */
        strcat(message_p->message, "{");

        /* convert port to string for data dump */
        itoa(message_p->id, output, 10);

        /* add id to the message */
        strcat(message_p->message, output);

        if(need_ack == SHUTDOWN)
        {
            /* repeat shutdown message 3 times */
            message_p->count = 3;
        }
        else
        {
            /* repeat normal message 10 times */
            message_p->count = 10;
        }
    }
    else
    {
        /* message without ack, repeat the message 1 time */
        message_p->count = 1;
    }

    /* add '\r' to the message */
    strcat(message_p->message, "\r");

    /*
     * Interval of 5 secondes before retry of the message (will be
     * doubled on each retry)
     */
    message_p->interval = 5;

    /*
     * Initial deposit is delayed if there are more messages, separation
     * of 10 seconde between subsequent messages is used.
     */
    /* start (10 * number) seconds from now */
    start_timer(&(message_p->timer), (10 * number) * TM_SEC);

    /* add to msg_list_p, new message in front */
    message_p->next = msg_list_p;
    msg_list_p = message_p;
}

/*
 * send an item beacon on request of a station
 *
 * Transmission is not done here, but the item is added to a
 * list of messages which are scheduled for transmission.
 * An item is transmitted two times.
 */
static void send_item(short number, char *message, uidata_t *uidata_p)
{
    message_t *message_p;

    if(keyboard_command == 1)
    {
        say("Object %d: %s\r\n", number, message);
        return;
    }

    /* create a new message */
    message_p = malloc(sizeof(message_t));

    if(message_p == NULL)
    {
        /* no memory, more luck next time! */
        return;
    }

    /* 
     * Copy message in. Maximum length of a message is = MAX_MESSAGE_LENGTH
     * bytes.
     */
    message_p->message[0] = '\0';
    strncpy(message_p->message, message, MAX_MESSAGE_LENGTH);

    /* make shure there is a null terminator at the end */
    message_p->message[MAX_MESSAGE_LENGTH] = '\0';

    /* send back where the message came from */
    message_p->port = uidata_p->port;

    /* no message id */
    message_p->id = -1;

    /* 
     * message without ack, repeat the message twice to get a more
     * reliable reception.
     */
    message_p->count = 2;

    /* add '\r' to the message */
    strcat(message_p->message, "\r");

    /*
     * Start in 10 secondes, the decay algorithm for non-messages will
     * repeat the object the second time 60 minutes after the first
     * deposit.
     */
    message_p->interval = 10 * number;

    /* start interval-time seconds from now */
    start_timer(&(message_p->timer), message_p->interval * TM_SEC);

    /* add to msg_list_p, new message in front */
    message_p->next = msg_list_p;
    msg_list_p = message_p;
}

/*
 * Find the first response to a questions, return default response when
 * a specific response can not be found. Return NULL when also a
 * default response can't be found.
 */
static char *query_first(char *question)
{
    char    *default_answer;
    query_t *query_p;
    char     query[MAX_MESSAGE_LENGTH + 1];
    char    *p;

    /* check if we need to re-read the message file */
    check_queries();

    default_answer = NULL;

    query_p = digi_query;

    while(query_p != NULL)
    {
        /*
         * Try to fine a match for this query
         */
        strcpy(query, query_p->query);

        p = strtok(query, "|");

        while(p != NULL)
        {
            if(*p == '?')
            {
                p++;
            }

            if(strcmp(p, question) == 0)
            {
                digi_query_cur_p = query_p->next;

                /* found! */
                return query_p->answer;
            }
            
            p = strtok(NULL, "|");
        }

        /*
         * Didn't find a match. If this is the default answer then
         * remember it in case we don't find a match at all
         */
        if(strcmp(query_p->query, "?") == 0)
        {
            /* found default answer, remember */
            default_answer = query_p->answer;
        }

        query_p = query_p->next;
    }

    digi_query_cur_p = NULL;

    /* if we end up here, we haven't found a match */
    return default_answer;
}

/*
 * Find the next response to a questions, return NULL when no more
 * specific responses can not be found.
 */
char *query_next(char *question)
{
    query_t *query_p;
    char     query[MAX_MESSAGE_LENGTH + 1];
    char    *p;

    query_p = digi_query_cur_p;

    while(query_p != NULL)
    {
        /*
         * Try to fine a match for this query
         */
        strcpy(query, query_p->query);

        p = strtok(query, "|");

        while(p != NULL)
        {
            if(*p == '?')
            {
                p++;
            }

            if(strcmp(p, question) == 0)
            {
                digi_query_cur_p = query_p->next;

                /* found! */
                return query_p->answer;
            }
            
            p = strtok(NULL, "|");
        }

        query_p = query_p->next;
    }

    digi_query_cur_p = NULL;

    return NULL;
}

static void send_autoreply_query(char *question, uidata_t *uidata_p)
{
    char *p;
    short number;

    number = 1;

    p = query_first(question);
    if(p == NULL)
    {
        send_message(ACK, number++, "No messages specified", uidata_p);
    }
    else
    {
        while(p != NULL)
        {
            if((*p == ';') || (*p == ')'))
            {
                send_item(number++, p, uidata_p);
            }
            else
            {
                send_message(ACK, number++, p, uidata_p);
            }
            p = query_next(question);
        }
    }
    return;
}

static void send_autoreply_mheard_port(short port, uidata_t *uidata_p)
{
    char      message[MAX_MESSAGE_LENGTH + 1];
    char      call[CALL_LENGTH + 1];
    mheard_t *mheard_p;
    short     count;
    short     number;

    if(port > mac_ports())
    {
        sprintf(message, "Digi port %d not active", port);
        send_message(ACK, 1, message, uidata_p);
        return;
    }
    else
    {
        sprintf(message, "Heard direct on port %d:", port);
    }

    mheard_p = first_mheard(port - 1); /* 0 based port numbers */
    if(mheard_p == NULL)
    {
        strcat(message, " none");
    }

    number = 1;

    send_message(ACK, number++, message, uidata_p);

    count = 0;

    /* Layout using worst-case length calls */
    /* THD7E: |Line numbr 1Line numbr 2Line numbr 3Line numb| */
    /*         P1 PE1DNN-12  PE1DNN-12  PE1DNN-12  PE1DNN-12  */

    /* start with port number */
    sprintf(message, "P%d ", port);

    while(mheard_p != NULL)
    {
        /* if there are already calls in the message then add 2 space */
        /* characters to separate the calls from eachother            */
        if(count > 0)
        {
            strcat(message, "  ");
        }

        /* increase count */
        count++;

        /* fill room of the size of a full call with spaces */
        /*           "PE1DNN-12"                            */
        strcpy(call, "         ");
        
        /* patch in the call from the mheard structure */
        strncpy(call, mheard_p->call, strlen(mheard_p->call));
        
        /* add this to the message */
        strcat(message, call);

        /* when there are 4 calls in the message, transmit */
        if(count == 4)
        {
            send_message(ACK, number++, message, uidata_p);

            /* start all over */
            count = 0;

            /* start with port number */
            sprintf(message, "P%d ", port);
        }

        mheard_p = next_mheard(port - 1); /* 0 based port numbers */
    }

    if(count != 0)
    {
        /* send remaining calls */
        send_message(ACK, number, message, uidata_p);
    }
}

static void send_autoreply_mheard_call(char *call, uidata_t *uidata_p)
{
    char       message[MAX_MESSAGE_LENGTH + 1];
    char       stamp[MAX_MESSAGE_LENGTH + 1];
    mheard_t  *mheard_p;
    struct tm *t;

    mheard_p = find_mheard(call);
    if(mheard_p == NULL)
    {
        sprintf(message, "Not found:  \"%s\"", call);
    }
    else
    {
        tzset();
        t = localtime(&(mheard_p->when));

        strftime(stamp, MAX_MESSAGE_LENGTH - CALL_LENGTH - 1 /* port-number */,
                        " rx %b %d %Y - %H:%M:%S on port ", t);

        sprintf(message, "%9s%s%d", mheard_p->call, stamp, mheard_p->port + 1);
    }

    send_message(ACK, 1, message, uidata_p);
}

static void send_autoreply_aprsd(uidata_t *uidata_p)
{
    char      message[MAX_MESSAGE_LENGTH + 1];
    mheard_t *mheard_p;
    short     count;

    strcpy(message, "Directs=");

    mheard_p = first_mheard(-1); /* -1 = on all ports */

    count = 0;
    while(mheard_p != NULL)
    {
        /* increase count */
        count++;

        /* catinate the call to the message with leading space */
        strcat(message, " ");
        strcat(message, mheard_p->call);
        
        /* when there are 5 calls in the message, transmit and quit */
        if(count == 5)
        {
            send_message(NOACK, 1, message, uidata_p);
            return;
        }

        mheard_p = next_mheard(-1);
    }

    if(count != 0)
    {
        /* send calls */
        send_message(NOACK, 1, message, uidata_p);
    }
}

static void send_autoreply_ping_aprst(uidata_t *uidata_p)
{
    char  data[MAX_DATA_LENGTH + 1];
    short i;

    /*
     * Worst-case length: 10 calls with full SSID = 10 * 9 = 90 bytes
     *                     1 time ">"                         1 byte
     *                     8 times ","                        8 bytes
     *                     1 times "*"                        1 byte
     *                     1 times ":"                        1 byte
     *                    grand total                       101 bytes
     *
     * This will fit into our data-space without problems
     */

    data[0] = '\0';

    strcat(data, uidata_p->originator);
    strcat(data,">");
    strcat(data, uidata_p->destination);

    for(i = 0; i < uidata_p->ndigi; i++)
    {
        strcat(data,",");
        strcat(data,uidata_p->digipeater[i]);

        if((uidata_p->digi_flags[i] & H_FLAG) != 0)
        {
            /* this one has the H_FLAG set */
            if(i == (uidata_p->ndigi -1))
            {
                /* this is the last digi, add asterix to this call */
                strcat(data,"*");
            }
            else
            {
                /*
                 * Not the last digi, look if the next one also has the
                 * H_FLAG set
                 */
                if((uidata_p->digi_flags[i + 1] & H_FLAG) == 0)
                {
                    /* next digi did not repeat, put asterix on this one */
                    strcat(data,"*");
                }
            }
        }
    }

    /* add a colon at the end */
    strcat(data, ":");

    /*
     * note: send_message will truncate the data to the allowed maximum
     * message length, the rest is lost
     */
    send_message(NOACK, 1, data, uidata_p);
}

static void send_autoreply_aprsm(uidata_t *uidata_p)
{
    message_t *message_p;
    char       message_prefix[MSG_PREFIX_LENGTH + 1];

    /* prepare a message prefix with the originator-call for comparison */
    /* prepare 2 colons with space for a full call ":PE1DNN-12:"        */
    strcpy(message_prefix, ":         :");

    /* patch in the originator-call */
    strncpy(&(message_prefix[1]), uidata_p->originator,
                                strlen(uidata_p->originator));
    
    /* loop though the messages to transmit */
    message_p = msg_list_p;
    while(message_p != NULL)
    {
        /* go through all unacked transmit messages */
        if(message_p->count > 0)
        {
            /* check if this is for the requester */
            if(strncmp(message_p->message, message_prefix, 
                                        MSG_PREFIX_LENGTH) == 0)
            {
                /* transmit_message */
                transmit_message(message_p);
            }
        }
        /* move to next */
        message_p = message_p->next;
    }
}

static void send_autoreply_date(uidata_t *uidata_p)
{
    char       message[MAX_MESSAGE_LENGTH + 1];
    time_t     curtime;
    struct tm *loctime;

    /* Make a timestamp */

    /* Determine current time and convert into local representation. */
    curtime = time (NULL);
    loctime = localtime (&curtime);

    /* Format the time */
    strftime(message, MAX_MESSAGE_LENGTH- 1, "%b %d %Y - %H:%M:%S", loctime);

    /* send date and time */
    send_message(ACK, 1, message, uidata_p);
}

static void send_autoreply_exit(uidata_t *uidata_p)
{
    send_message(SHUTDOWN, 1, "DIGI_NED: System shutdown after acknowlege",
                               uidata_p);
}

static void send_autoreply(char *bp, uidata_t *uidata_p)
{
    char  *p;

    if(keyboard_command == 0)
    {
        vsay("New message, send auto-response message back\r\n");
    }

    /* start at the beginning */
    p = bp;

    if(*p == '?')
    {
        /* skip starting '?' if the query starts with '?' */
        p++;
    }

    if((strcmp(p, "ping?") == 0) || (strcmp(p, "aprst") == 0))
    {
        send_autoreply_ping_aprst(uidata_p);
        return;
    }
    else if(strcmp(p, "aprsd") == 0)
    {
        send_autoreply_aprsd(uidata_p);
        return;
    }
    else if(strcmp(p, "aprsm") == 0)
    {
        send_autoreply_aprsm(uidata_p);
        return;
    }
    else if(strcmp(p, "aprs?") == 0)
    {
        send_all();
        return;
    }
    else if(strcmp(p, "date") == 0)
    {
        send_autoreply_date(uidata_p);
        return;
    }
    else if(strncmp(p, "mheard ", 7) == 0)
    {
        if( (strcmp(p, "mheard <port>") == 0)
            ||
            (strcmp(p, "mheard <call>") == 0)
          )
        {
            send_autoreply_query(p, uidata_p);
            return;
        }
        else if(strlen(p) == 8)
        {
            if((p[7] > '0') && (p[7] < '9'))
            {
                send_autoreply_mheard_port(atoi(&p[7]), uidata_p);
                return;
            }
            else
            {
                send_autoreply_query("?mheard <port>", uidata_p);
                return;
            }
        }
        else
        {
            if(strlen(&p[7]) < 10)
            {
                send_autoreply_mheard_call(&p[7], uidata_p);
                return;
            }
            else
            {
                send_autoreply_query("?mheard <call>", uidata_p);
                return;
            }
        }
    }
    else if(strncmp(p, "mh ", 3) == 0)
    {
        if( (strcmp(p, "mh <port>") == 0)
            ||
            (strcmp(p, "mh <call>") == 0)
          )
        {
            send_autoreply_query(p, uidata_p);
            return;
        }
        else if(strlen(p) == 4)
        {
            if((p[3] > '0') && (p[3] < '9'))
            {
                send_autoreply_mheard_port(atoi(&p[3]), uidata_p);
                return;
            }
            else
            {
                send_autoreply_query("?mh <port>", uidata_p);
                return;
            }
        }
        else
        {
            if(strlen(&p[3]) < 10)
            {
                send_autoreply_mheard_call(&p[3], uidata_p);
                return;
            }
            else
            {
                send_autoreply_query("?mh <call>", uidata_p);
                return;
            }
        }
    }
    else if(strncmp(p, "exit", 4) == 0)
    {
        if( (is_owner(uidata_p->originator) == 1)
            &&
            (digi_enable_exit == 1)
          )
        {
            if(strncmp(p, "exit ", 5) == 0)
            {
                exit_info.exitcode = atoi(&p[5]) & 0x00ff;

                vsay("Recieved shutdown request by digi owner's remote "
                    "\"?exit %d\" command\r\n", exit_info.exitcode);
            }
            else
            {
                exit_info.exitcode = 2;

                vsay("Recieved shutdown request by digi owner's remote "
                    "\"?exit\" command\r\n");
            }
            send_autoreply_exit(uidata_p);
            return;
        }
    }

    send_autoreply_query(p, uidata_p);
}

static void parse_message(char *bp, uidata_t *uidata_p)
{
    char      *p;
    short      id;
    short      new;
    char       stamp[MAX_MESSAGE_LENGTH + 1];
    time_t     curtime;
    struct tm *loctime;

    /* Make a timestamp */

    /* Determine current time and convert into local representation. */
    curtime = time (NULL);
    loctime = localtime (&curtime);

    /* Format the time */
    strftime(stamp, MAX_MESSAGE_LENGTH- 1, "%b %d %Y - %H:%M:%S", loctime);

    /* strip '\r' from message and truncate there  */
    p = strrchr(bp, '\r');
    if(p != NULL)
    {
        *p = '\0';
    }

    vsay("\r\n===========\r\n");
    vsay("Got message: \"%s\" from %s at %s\r\n",
                               bp, uidata_p->originator, stamp);

    p = strrchr(bp, '{');
    if(p == NULL)
    {
        /* No sequence number. Check if it is and 'ack' or 'rej'     */
        /* On rej also do not retransmit, the receiver is overloaded */
        if( (strncmp("ack", bp, 3) == 0)
            ||
            (strncmp("rej", bp, 3) == 0)
          )
        {
            id = atoi(&bp[3]);

            /* remove message */
            remove_message(id);

            /* 
             * check if we are closing down and if this is the ack
             * the close down procedure was waiting for
             */
            if(exit_info.state == CLOSING_DOWN)
            {
                if(id == exit_info.id)
                {
                    /* yes, shutdown acknoledged */
                    vsay("Exiting on %s...\r\n", stamp);
                    exit(exit_info.exitcode);
                }
            }

            /* for an ACK we are finished now */
            return;
        }
    }
    else
    {
        /* truncate the message and retrieve the sequence number */

        /* truncate message at this point */
        *p = '\0';
        /* advance to just after the '{' */
        p++;

        send_ack(p, uidata_p);
    }

    /*
     * Walk back from the end to the beginning and remove all trailing
     * spaces and tabs
     */
    p = strrchr(bp, '\0');
    while(p != bp)
    {
        /* post decrement */
        p--;

        if((*p == ' ') || (*p == '\t'))
        {
            *p = '\0';
        }
        else
        {
            /* not a space or tab, break out of the loop */
            break;
        }
    }

    /* convert to lowercase */
    strlwr(bp);

    new = is_new_message(bp, uidata_p);

    /* take only messages which are new for this sender */
    if(new == 1)
    {
        /* new message, send auto-response message back */
        send_autoreply(bp, uidata_p);

        /* remember that we already got this message */
        keep_old_message(bp, uidata_p);
    }
}

void do_messages(void)
{
    message_t *message_p;
    message_t *next_message_p;
    message_t *previous_message_p;

    /* first do the transmissions, then cleanup */

    /* loop though the messages to see which ones are due to transmit */
    message_p = msg_list_p;
    while(message_p != NULL)
    {
        /* transmit if the timer for this message is elapsed */
        if(is_elapsed(&(message_p->timer)))
        {
            /* time to transmit the message */
            if(message_p->count > 0)
            {
                /* transmit_message */
                transmit_message(message_p);

                /* lower the count */
                message_p->count--;

                /*
                 * we are finished with transmission of this message,
                 * display visual separator.
                 */
                display_separator();

                if((message_p->message[0] == ';')
                   || 
                   (message_p->message[0] == ')'))
                {
                    /*
                     * For objects and items:
                     * fixed time of 60 seconds before retransmission
                     */
                    message_p->interval = 60 * TM_SEC;
                }
                else
                {
                    /*
                     * For messages:
                     * every time double the time before retransmission
                     */
                    message_p->interval *= 2;
                }

                /* 
                 * setup the timer
                 */
                start_timer(&(message_p->timer), message_p->interval);
            }
        }
        /* move to next */
        message_p = message_p->next;
    }

    /*
     * loop though messages to see which ones can be removed
     *
     * A message can be removed if the count has either reached
     * zero or is set to zero upon reception of an ack. The timer
     * must be expired so that there was enough time to acknowledge
     * the last transmission (this important when an action is
     * triggered by the ack, for example shutdown).
     */
    message_p          = msg_list_p;
    previous_message_p = NULL;
    while(message_p != NULL)
    {
        /* save next */
        next_message_p = message_p->next;
        
        /* 
         * if not for repetition anymore and time to receive an ack
         * is passed then remove the message
         */
        if((message_p->count == 0) && (is_elapsed(&(message_p->timer))))
        {
            /* 
             * if the shutdown messgae is about to be destroyed
             * go back to normal operation (shutdown was not
             * acknowledged)
             */
            if(exit_info.state == CLOSING_DOWN)
            {
                if(message_p->id == exit_info.id)
                {
                    /* shutdown message not acknowledged */
                    vsay("Shutdown aborted, "
                        "switched back to active operation\r\n");
                    exit_info.state = ACTIVE;
                }
            }

            /* destroy this message */
            if(previous_message_p == NULL)
            {
                /* still at the beginning of the list */
                msg_list_p = next_message_p;
            }
            else
            {
                /* not at the beginning of the list */
                previous_message_p->next = next_message_p;
            }

            /* if an ID is used free it */
            if(message_p->id != -1)
            {
                message_id[(short) message_p->id] = 0;
            }

            /* throw away this message */
            free(message_p);

            /* debugging, dump outstanding message ids */
            {
                short i;
                vsay("Outstanding msg-id's: ");
                for(i = 0; i < MAX_NR_OF_IDS; i++)
                {
                    if(message_id[i] != 0)
                    {
                        vsay("%d,", i);
                    }
                }
                vsay("\r\n");
            }
            display_separator();

            /* let message_p point to next message */
            message_p = next_message_p;
        }
        else
        {
            /* keep this message */
            previous_message_p = message_p;
            message_p = next_message_p;
        }
    }
}

short recv_message(uidata_t *uidata_p)
{
    short res;
    char *bp;
    char  data[MSG_PREFIX_LENGTH + MAX_MESSAGE_LENGTH + MSG_SUFFIX_LENGTH + 1];
    char  message_prefix[MSG_PREFIX_LENGTH + 1];

    /* Find out if this is a message and to whom it is addressed */

    /* Minimum data-content: ":PE1DNN-12:{1", 13 bytes */
    if(uidata_p->size < 13)
    {
        /* can't be a message */
        return 0;
    }

    /* don't handle too big data packets */
    if(uidata_p->size > 
       (MSG_PREFIX_LENGTH + MAX_MESSAGE_LENGTH + MSG_SUFFIX_LENGTH)
      )
    {
        /* can't be a message */
        return 0;
    }

    /* convert received data to string */
    memcpy(data, uidata_p->data, uidata_p->size);
    data[uidata_p->size] = '\0';

    /* let bp point to the start of the string */
    bp = &data[0];

    /*
     * First check if this is repeated through a tunnel.
     * Third-party messages can be recursive so we have
     * to dig through until we reach the last one before
     * the message. That will be the real originator.
     */
    while(*bp == '}')
    {
        /* Make the sender of the message the originator */

        /* Try to find the end of the call */
        bp = strchr(bp, '>');
        if(bp == NULL)
        {
            /* Original originator missing */
            return 0;
        }
        /* Make this a string by patching the '>' character to '\0' */
        *bp = '\0';

        /* check length of the real originator */
        if(strlen(&data[1]) >= CALL_LENGTH)
        {
            /* Original originator too long */
            return 0;
        }
        
        /*
         * Overwrite the originator of the packet with the call
         * recovered from the tunnel data
         */
        strcpy(uidata_p->originator, &data[1]); 

        /* advance bp after the '\0' character we put in  */
        bp++;

        /* Move to the end of the 3rd party path */
        bp = strchr(bp, ':');
        if(bp == NULL)
        {
            /* can't be a message */
            return 0;
        }
        /* move to the first char of the data */
        bp++;
    }

    /* prepare a message prefix with the digi-call for comparison */
    /* prepare 2 colons with space for a full call ":PE1DNN-12:"  */
    strcpy(message_prefix, ":         :");

    /* patch in the digi-call */
    strncpy(&(message_prefix[1]), digi_digi_call, strlen(digi_digi_call));

    if(strncmp(bp, message_prefix, MSG_PREFIX_LENGTH) == 0)
    {
        /* Check to see if originator is on the message blocked list. */
        res = is_msg_block(uidata_p->originator);

        if(res == 1)
        {
            vsay("\r\n===========\r\nMessage Check: ");
            vsay("Originator blocked: ignore message\r\n");
            /* it was a message to addressed to me */
            return 1;
        }

        if(res == -1)
        {
            vsay("\r\n===========\r\nMessage Check: ");
            vsay("I will not react on my own message\r\n");
            /* it was a message to addressed to me */
            return 1;
        }

        if (number_of_hops(uidata_p) > digi_max_msg_hops)
        {
            /* maximum number of hops exceeded */
            vsay("\r\n==========\r\nMessage Check:\n");
            vsay("Maximum number of %d hop(s) exceeded.\n", digi_max_msg_hops);
            vsay("Message ingnored.\n");
            return 1;
        }

        /* move bp passed the message_prefix */
        bp = &(bp[MSG_PREFIX_LENGTH]);

        parse_message(bp, uidata_p);

        /* it was a message to addressed to me */
        return 1;
    }

    /* it was not a message for me */
    return 0;
}

void keyboard_message(char *command)
{
    uidata_t uidata;

    /* fill in a fake uidata frame */

    uidata.port = 0;
    uidata.distance = LOCAL;
    uidata.primitive = 0x03; /* UI */
    uidata.pid = 0xFC;
    strcpy(uidata.originator,"KEYBOARD");
    uidata.orig_flags = 0;
    strcpy(uidata.destination,"SCREEN");
    uidata.dest_flags = 0;
    uidata.ndigi = 0;
    strcpy(uidata.data,":KEYBOARD :");
    strcat(uidata.data, command);
    uidata.size = strlen(uidata.data);

    /* keyboard command */
    keyboard_command = 1;

    /* convert command to lowercase */
    strlwr(command);

    /* call handler */
    send_autoreply(command, &uidata);

    /* not a keyboard command */
    keyboard_command = 0;
}

void init_message(void)
{
    short i;

    for(i = 0; i < MAX_NR_OF_IDS; i++)
    {
        message_id[i] = 0;
    }

    /* for boot-time message */
    boot_time = time(0);

    /* read the message query/answer file */
    read_queries();

    /* not a keyboard command */
    keyboard_command = 0;

    /* active operation, no shutdown in progress */
    exit_info.state = ACTIVE;

    msg_list_p = NULL;
}
