/*
 *  Copyright (C) 2000  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _LINUX_
#include "linux.h"
#endif /* _LINUX_ */
#include "main.h"
#include "digicons.h"
#include "digitype.h"
#include "genlib.h"
#include "mac_if.h"
#include "read_ini.h"
#include "send.h"
#include "keep.h"
#include "timer.h"
#include "output.h"

/*
 * This file contains all code to interpret and execute 'send' commands
 * from the .ini file.
 *
 * Example:
 * send: 20 all DIGI_DEST,WIDE,TRACE6-6\r\n
 */

/*
 * Linked list containing the 'send' command data
 */
static send_t *digi_send_p = NULL;

/*
 * Add send command from the .ini file to the internal administration.
 *
 * send: 20 all DIGI_DEST,WIDE,WIDE,TRACE6-6\r\n
 *      ^
 *
 * On entry it is assumed that 'strtok' is tokenizing the line and the
 * 'send' token is already handled. The tokenizer is now at the point in
 * the string indicated by '^' above (the ':' was changed to '\0' by
 * 'strtok').
 */
void add_send(FILE *fd, short ini_ports)
{
    send_t send;
    send_t *send_p;
    char   *tmp;
    char   *p;
    char   *path;
    char    output[OUTPORTS_LENGTH];
    char    number[MAX_ITOA_LENGTH];
    char    line[256];
    short   i;
    short   port;

    tmp = strtok(NULL," \t");

    send.interval = atoi(tmp);
    if(send.interval == 0)
    {
        say("Error in interval in send: at line %d, number > 0 expected\r\n", 
                 linecount);
        return;
    }

    tmp = strtok(NULL, " \t\r\n");
    if(tmp == NULL)
    {
        say("Missing to-port on send: at line %d\r\n", linecount);
        return;
    }

    if(check_ports(tmp, "To-port", "send", linecount, EXCL_ALLBUT) == 0)
    {
        return;
    }

    strcpy(send.output, tmp);

    send.path = NULL;

    path = strtok(NULL, "\r\n");

    if(path != NULL)
    {
        strupr(path);

        send.path = strdup(path);

        if(send.path == NULL)
        {
            say("Out of memory on send: at line %d\r\n", linecount);
            return;
        }

        /*
         * Check for illegal digipeaters (too long)
         */
        tmp = strtok(path, " ,\t\r\n");
        while(tmp != NULL)
        {
            if(strlen(tmp) > 9)
            {
                say("Argument (%s) to long on send: at line %d\r\n",
                        tmp, linecount);

                free(send.path);
                return;
            }
            tmp = strtok(NULL, " ,\t\r\n");
        }
    }

    /*
     * convert output shortcuts to full-lists
     * while we are at it, convert to 0 base portnumbers
     */
    if(strcmp(send.output,"all") == 0)
    {
        send.output[0] = '\0';
        for(i = 0; i < ini_ports; i++)
        {
            if(send.output[0] != '\0')
                strcat(send.output, ",");
            strcat(send.output, itoa(i, number, 10));
        }
    }
    else
    {
        /* start with empty output string, add the digits we find */
        output[0] = '\0';
        tmp = strtok(send.output, ",");
        while(tmp != NULL)
        {
            port = atoi(tmp);
            if(port < 1 || port > ini_ports)
            {
                say("To-port out of range (1-%d) on send: at line %d\r\n",
                        ini_ports, linecount);
                return;
            }
            port--; /* make base 0 numbers */

            if(output[0] != '\0')
                strcat(output, ",");
            strcat(output, itoa(port, number, 10));

            tmp = strtok(NULL, ",");
        }
        strcpy(send.output, output);
    }

    if(fgets(&line[0], 255, fd) != NULL)
    {
        /* slash handling */
        tmp = &line[0];
        while(*tmp != '\0')
        {
#ifndef _LINUX_
            if(*tmp == '/' ) { *tmp = '\\'; }
#else
            if(*tmp == '\\') { *tmp = '/';  }
#endif /* _LINUX_ */
            tmp++;
        }

        /* strip newline from filename */
        tmp = strtok(line, "\r\n");

        /* see if we have to add progpath to the filename */
        if( (strrchr(tmp, (int) '/') == NULL)
            &&
            (strrchr(tmp, (int) '\\') == NULL)
            &&
            (strrchr(tmp, (int) ':') == NULL)
          )
        {
            /* no path in name, add progpath in front  */
            p = malloc(strlen(get_progpath()) + strlen(tmp) + 1);
            if(p != NULL)
            {
                strcpy(p, get_progpath());
                strcat(p, tmp);
            }
            send.file = p;
        }
        else
        {
            /* name includes path */
            send.file = strdup(tmp);
        }

        if(send.file == NULL)
        {
            if(send.path != NULL)
                free(send.path);
            say("Out of memory on send: at line %d\r\n", linecount);
            return;
        }
        linecount++;
    }
    else
    {
        say("Early end-of-file on send: at line %d\r\n", linecount);
        if(send.path != NULL)
            free(send.path);
        return;
    }

    send_p = (send_t*) malloc(sizeof(send_t));

    /*
     * if out of memory then ignore
     */
    if(send_p == NULL)
    {
        say("Out of memory while handling send: line at line %d\r\n",
                linecount);
        return;
    }
    
    /* initialize the timer, start with transmission after 15 seconds */
    start_timer(&send.timer, 15);

    /*
     * Duplicate data
     */
    memcpy(send_p, &send, sizeof(send_t));

    /*
     * Add to sender-list
     */
    send_p->next = digi_send_p;
    digi_send_p = send_p;
}

/*
 * Transmit the data specified by the file which is specified 
 * in the send_t structure.
 */
static void transmit_send(send_t *send_p)
{
    uidata_t uidata;
    short    ndigi;
    char    *tmp;
    char    *tmp2;
    char    *path;
    char     output[OUTPORTS_LENGTH];
    char     line[256];
    FILE    *fd;
    short    n;

    /* prepare transmission */
    uidata.port = 0;

    /* transmit remotely */
    uidata.distance = REMOTE;

    /* primitive UI with no poll bit */
    uidata.primitive = 0x03;

    /* PID = 0xF0, normal AX25*/
    uidata.pid = 0xF0;

    /* fill in originator */
    strcpy(uidata.originator, digi_digi_call);
    uidata.orig_flags = (char) (RR_FLAG | C_FLAG);

    /* fill in destination */
    strcpy(uidata.destination, digi_digi_dest);
    uidata.dest_flags = RR_FLAG;

    /* make copy of digipeater path for strtok */
    ndigi=0;
    path = strdup(send_p->path);
    if(path == NULL)
    {
        /* no memory for transmission */
        return;
    }

    /* overwrite destination if it is in the path (first argument) */
    tmp = strtok(path, " ,\t\r\n");
    if(tmp != NULL)
    {
        /* only need to set destination, dest_flags value was set above */
        strcpy(uidata.destination, tmp);
    }

    /* add digipeater path */
    tmp = strtok(NULL, " ,\t\r\n");
    while(tmp != NULL)
    {
        /* if there are more that MAX_DIGI digi's we ignore the rest */
        if(ndigi < MAX_DIGI)
        {
            strcpy(uidata.digipeater[ndigi], tmp);
            uidata.digi_flags[ndigi] = RR_FLAG;
            ndigi++;
        }
        tmp = strtok(NULL, " ,\t\r\n");
    }
    uidata.ndigi = ndigi;

    /* don't need "path" anymore, free memory */
    free(path);

    /* get the transmission data from file */
    fd = fopen(send_p->file, "rt");
    if(fd == NULL)
    {
        vsay("Can't open file \"%s\"\r\n", send_p->file);
        /* can't open file, nothing to transmit! */
        return;
    }

    n = fread(line, 1, 254, fd); /* use 254 to leave room to add a '\r' */
    if(n == -1)
    {
        /* can't read data, nothing to transmit! */
        fclose(fd);
        return;
    }
    line[n] = '\0';

    /* don't need file anymore */
    fclose(fd);

    /* remove away '\r' characters */
    tmp  = line;
    tmp2 = line;
    while(*tmp != '\0')
    {
        if(*tmp != '\r')
        {
            *tmp2 = *tmp;
            tmp2++;
        }
        tmp++;
    }
    *tmp2 = '\0';

    /* convert newline to cariage-return characters */
    tmp = line;
    while(*tmp != '\0')
    {
        if(*tmp == '\n')
        {
            *tmp = '\r';
        }
        tmp++;
    }

    /* check if there is a '\r' at the end, if not then add it */
    n = strlen(line);
    if((n == 0) || (line[n - 1] != '\r'))
    {
        /* add a <CR> at the end */
        strcat(line,"\r");
    }

    strcpy(uidata.data, line);
    uidata.size = strlen(uidata.data);

    /* avoid digipeating my own data */
    keep_old_data(&uidata);

    /* make a copy to be handled by strtok */
    strcpy(output, send_p->output);

    /* show what we will transmit */
    dump_uidata_to(output, &uidata);

    /* now do the transmission for each port */
    tmp = strtok(output, ",");
    while(tmp != NULL)
    {
        /* fill in the port */
        uidata.port = atoi(tmp);

        /* transmit */
        put_uidata(&uidata);

        /* get next port to transmit on */
        tmp = strtok(NULL, ",");
    }
}

/*
 * Send all beacons. Used for the "?aprs?" broadcast handling.
 * note that the timers are not restarted.
 */
void send_all()
{
    send_t     *send_p;

    /*
     * walk through the transmission list and start transmission.
     */
    send_p = digi_send_p;

    /*
     * While there are still send commands to consider
     */
    while(send_p != NULL)
    {
        /* it is time to send */
        transmit_send(send_p);

        /* separator between all transmissions, not after the last */
        if(send_p->next != NULL)
        {
            /* command handled, show separator */
            display_separator();
        }

        /* advanve to next send command */
        send_p = send_p->next;
    }
}

/*
 * Do send command, see if there is any transmission to be done at this
 * time, if there is do the transmission and restart the timer for the
 * next time.
 */
void do_send()
{
    send_t     *send_p;

    /*
     * walk through the transmission list and start transmission if
     * needed
     */
    send_p = digi_send_p;

    /*
     * While there are still send commands to consider
     */
    while(send_p != NULL)
    {
        if(is_elapsed(&(send_p->timer)) != 0)
        {
            /* it is time to send */
            transmit_send(send_p);

            /* command handled, show separator */
            display_separator();

            /* restart the timer */
            start_timer(&(send_p->timer), send_p->interval * TM_MIN);
        }

        /* advanve to next send command */
        send_p = send_p->next;
    }
}

/*
 * Swap occurences of DIGI_CALL and DIGI_DEST with the calls
 * specified in digi_digi_call and digi_digi_dest.
 *
 * This is done with a seperate routine since at the time the rules
 * were put in the list structure "digi_digi_call" and "digi_digi_dest"
 * did not need to have a defined value. When we call this routine
 * this is the case and the calls have been checked (on size).
 */

void swap_digicall_digidest_send()
{
    send_t     *send_p;
    char        path[MAX_DATA_LENGTH];
    char        oldpath[MAX_DATA_LENGTH];
    char       *call;
    short       replaced;

    /*
     * walk through the send list, replace DIGI_CALL and DIGI_DEST
     */
    send_p = digi_send_p;
    while(send_p != NULL)
    {
        replaced = 0;

        path[0] = '\0';
        /* make a copy, don't run tokenizer on original string */
        strcpy(oldpath, send_p->path);
        call = strtok(oldpath, " ,\t\r\n");
        while(call != NULL)
        {
            if(path[0] != '\0')
                strcat(path, ",");

            if(strcmp(call, "DIGI_CALL") == 0)
            {
                strcat(path, digi_digi_call);
                replaced = 1;
            }
            else if(strcmp(call, "DIGI_DEST") == 0)
            {
                strcat(path, digi_digi_dest);
                replaced = 1;
            }
            else
            {
                strcat(path, call);
            }
            call = strtok(NULL, " ,\t\r\n");
        }

        if(replaced == 1)
        {
            /*
             * !we can just overwrite since DIGI_CALL and DIGI_DEST
             * have the worse-case length!! (PE1DNN-15 and DIGI_CALL
             * have the same length!
             */
            strcpy(send_p->path, path);
        }

        send_p = send_p->next;
    }
}
